/*---------------------------------------------------------------------------
 * Copyright (C) 2008-2010 - Emanuele Bovisio
 *
 * This file is part of Mulk.
 *
 * Mulk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Mulk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Mulk.  If not, see <http://www.gnu.org/licenses/>.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 *---------------------------------------------------------------------------*/

#ifdef _WIN32
#include <direct.h>
#include <io.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <share.h>
#else
#include <unistd.h>
#endif
#include "file_obj.h"
#include "string_obj.h"
#include "url_list.h"
#include "option_obj.h"

#define MAX_LINE_LENGTH 1000
#define BUFFER_SIZE 1024


int copy(const char *from_filename, const char *to_filename)
{
	char buffer[BUFFER_SIZE];
	FILE *fp_to, *fp_from;
	int ret = -1;
	size_t count;

	if ((fp_from = fopen(from_filename, "rb")) == NULL)
		return -1;

	if ((fp_to = fopen(to_filename, "wb")) == NULL)
		goto Exit1;

	while ((count = fread(buffer, 1, BUFFER_SIZE, fp_from)) != 0)
		if (fwrite(buffer, 1, count, fp_to) < count)
			goto Exit;

	ret = (ferror(fp_to) || ferror(fp_from)) ? -1 : 0;

Exit:
	fclose(fp_to);
Exit1:
	fclose(fp_from);

	return ret;
}

size_t fwrite_offset(unsigned char *buf, size_t size, size_t nmemb, off_t offset, FILE *file)
{
	fseek(file, offset, SEEK_SET); 

	return fwrite(buf, size, nmemb, file);
}

int read_uri_from_text_file(const char *filename)
{
	FILE *file;
	char buf[MAX_LINE_LENGTH];

	if ((file = fopen(filename, "r")) == NULL)
		return 1;

	for (;;) {
		if (!fgets(buf, MAX_LINE_LENGTH, file))
			break;
		if (feof(file))
			break;

		string_trim(buf);

		/* skip empty lines and comments */
		if (!strlen(buf) || *buf == '#')
			continue;
 
		MULK_NOTE((_("Add url to download coming from text file: %s\n"), buf));
		add_new_url(buf);
	}

	fclose(file);

	return 0;
}

/* read only long option */
int read_option_from_text_file(const char *filename)
{
	FILE *file;
	char buf[MAX_LINE_LENGTH], *option, *value;
	int ret = 0;

	if ((file = fopen(filename, "r")) == NULL)
		return 1;

	for (;;) {
		if (!fgets(buf, MAX_LINE_LENGTH, file))
			break;
		if (feof(file))
			break;

		string_trim(buf);

		/* skip empty lines and comments */
		if (!strlen(buf) || *buf == '#')
			continue;

		option = buf;
		if ((value = strchr(buf, '=')) != NULL) {
			*value = 0;
			value++;
			if (!*value)
				value = NULL;
		}
		else
			value = NULL;

		MULK_NOTE((_("Add option coming from text file: %s%s%s\n"), option,
			value ? "=" : "", value ? value : ""));
	
		if ((ret = set_option(find_long_option(option), value)) != 0)
			break;
	}

	fclose(file);

	return ret;
}

#ifdef _WIN32

int make_dir_pathname(const char *pathname)
{
	struct _stat stat_buf;
	char *slash_ptr;
	char *path = NULL;

	if (!pathname || !*pathname)
		return 0;

	path = string_new(pathname);

	slash_ptr = strchr(path, *DIR_SEPAR_STR);

	while (slash_ptr) {
		*slash_ptr = 0;

		if (_stat(path, &stat_buf) < 0) {
			if (_mkdir(path) < 0) {
				string_free(&path);
				return 0;
			}
		}
		else if (!(stat_buf.st_mode & _S_IFDIR)) {
			string_free(&path);
			return 0;
		}

		*slash_ptr = *DIR_SEPAR_STR;
		slash_ptr = strchr(slash_ptr + 1, *DIR_SEPAR_STR);
	}

	string_free(&path);
	return 1;
}

int is_file_exist(const char *filename)
{
	struct _stat sts;

	if (_stat(filename, &sts) == -1 && errno == ENOENT)
		return 0;
	return 1;
}

int create_truncated_file(const char *filename, off_t size)
{
	int file, result = 1;

	if (!_sopen_s(&file, filename, _O_RDWR | _O_CREAT, _SH_DENYNO, _S_IREAD | _S_IWRITE)) {
		result = _chsize(file, size);
		_close(file);
	}

	return !result;
}

#else /* not _WIN32 */

int make_dir_pathname(const char *pathname)
{
	struct stat stat_buf;
	char *slash_ptr;
	char *path = NULL;
	int arg_mode = DEFAULT_MASK;

	if (!pathname || !*pathname)
		return 0;

	path = string_new(pathname);

	arg_mode |= (S_IWUSR | S_IXUSR);

	slash_ptr = strchr(path, *DIR_SEPAR_STR);

	while (slash_ptr) {
		*slash_ptr = 0;

		if (stat(path, &stat_buf) < 0) {
			if (mkdir(path, arg_mode) < 0) {
				string_free(&path);
				return 0;
			}
		}
		else if (!S_ISDIR(stat_buf.st_mode)) {
			string_free(&path);
			return 0;
		}

		*slash_ptr = *DIR_SEPAR_STR;
		slash_ptr = strchr(slash_ptr + 1, *DIR_SEPAR_STR);
	}

	string_free(&path);
	return 1;
}

int is_file_exist(const char *filename)
{
	struct stat sts;

	if (stat(filename, &sts) == -1 && errno == ENOENT)
		return 0;
	return 1;
}

int create_truncated_file(const char *filename, off_t size)
{
	if (truncate(filename, size))
		return 0;
	return 1;
}

#endif /* not _WIN32 */
