/*---------------------------------------------------------------------------
 * Copyright (C) 2008-2010 - Emanuele Bovisio
 *
 * This file is part of Mulk.
 *
 * Mulk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Mulk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Mulk.  If not, see <http://www.gnu.org/licenses/>.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 *---------------------------------------------------------------------------*/

/* Wrappers around malloc and memory debugging support.
   Copyright (C) 2003, 2004, 2005, 2006, 2007 Free Software Foundation,
   Inc.

This file is part of GNU Wget.

GNU Wget is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

GNU Wget is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wget.  If not, see <http://www.gnu.org/licenses/>.

In addition, as a special exception, the Free Software Foundation
gives permission to link the code of its release of Wget with the
OpenSSL project's "OpenSSL" library (or with modified versions of it
that use the same license as the "OpenSSL" library), and distribute
the linked executables.  You must obey the GNU General Public License
in all respects for all of the code used other than "OpenSSL".  If you
modify this file, you may extend this exception to your version of the
file, but you are not obligated to do so.  If you do not wish to do
so, delete this exception statement from your version.  */

/* modified and adapted to Mulk project by Emanuele Bovisio */

#include "m_malloc.h"

#if defined(DEBUG) || defined(_DEBUG)

#ifdef HAVE_STDINT_H
#include <stdint.h>  /* for uintptr_t */
#endif

#if defined( _MSC_VER ) && !defined( __cplusplus )
# define inline __inline
#endif // defined( _MSC_VER ) && !defined( __cplusplus )

#define SZ 100003               /* Prime just over 100,000.  Increase it to debug larger runs.  */
#define PTR_FORMAT(p) (int) (2 * sizeof (void *)), (unsigned long) (p)

static int malloc_count, free_count;

/* Hashing of numeric values, such as pointers and integers.

   This implementation is the Robert Jenkins' 32 bit Mix Function,
   with a simple adaptation for 64-bit values.  According to Jenkins
   it should offer excellent spreading of values.  Unlike the popular
   Knuth's multiplication hash, this function doesn't need to know the
   hash table size to work.  */
static unsigned long hash_pointer (const void *ptr)
{
  uintptr_t key = (uintptr_t) ptr;
  key += (key << 12);
  key ^= (key >> 22);
  key += (key << 4);
  key ^= (key >> 9);
  key += (key << 10);
  key ^= (key >> 2);
  key += (key << 7);
  key ^= (key >> 12);
#if SIZEOF_VOID_P > 4
  key += (key << 44);
  key ^= (key >> 54);
  key += (key << 36);
  key ^= (key >> 41);
  key += (key << 42);
  key ^= (key >> 34);
  key += (key << 39);
  key ^= (key >> 44);
#endif
  return (unsigned long) key;
}

static struct {
  const void *ptr;
  const char *file;
  int line;
} malloc_table[SZ];

/* Find PTR's position in malloc_table.  If PTR is not found, return
   the next available position.  */
static inline int ptr_position(const void *ptr)
{
	int i = hash_pointer(ptr) % SZ;
	for (; malloc_table[i].ptr != NULL; i = (i + 1) % SZ)
		if (malloc_table[i].ptr == ptr)
			return i;
	return i;
}

/* Register PTR in malloc_table.  Abort if this is not possible
   (presumably due to the number of current allocations exceeding the
   size of malloc_table.)  */
static void register_ptr(const void *ptr, const char *file, int line)
{
	int i;

	if (malloc_count - free_count > SZ) {
		fprintf(stderr, _("ERROR: increase SZ to a larger value and recompile\n"));
		exit(EXIT_FAILURE);
	}

	i = ptr_position(ptr);
	malloc_table[i].ptr = ptr;
	malloc_table[i].file = file;
	malloc_table[i].line = line;
}

/* Unregister PTR from malloc_table.  Return false if PTR is not
   present in malloc_table.  */
static int unregister_ptr(const void *ptr)
{
	int i = ptr_position(ptr);
	if (malloc_table[i].ptr == NULL)
		return 0;
	malloc_table[i].ptr = NULL;

	/* Relocate malloc_table entries immediately following PTR. */
	for (i = (i + 1) % SZ; malloc_table[i].ptr != NULL; i = (i + 1) % SZ) {
		const void *ptr2 = malloc_table[i].ptr;
		/* Find the new location for the key. */
		int j = hash_pointer(ptr2) % SZ;
		for (; malloc_table[j].ptr != NULL; j = (j + 1) % SZ)
			if (ptr2 == malloc_table[j].ptr)
				/* No need to relocate entry at [i]; it's already at or near
				   its hash position. */
				goto cont_outer;
		malloc_table[j] = malloc_table[i];
		malloc_table[i].ptr = NULL;
cont_outer:
		;
	}
	return 1;
}

/* Print the malloc debug stats gathered from the above information.
   Currently this is the count of mallocs, frees, the difference
   between the two, and the dump of the contents of malloc_table.  The
   last part are the memory leaks.  */
void print_malloc_debug_stats(void)
{
	int i;

	printf(_("\nMalloc:  %d\nFree:    %d\nBalance: %d\n\n"),
		malloc_count, free_count, malloc_count - free_count);
	for (i = 0; i < SZ; i++)
		if (malloc_table[i].ptr != NULL)
			printf("0x%0*lx: %s:%d\n", PTR_FORMAT (malloc_table[i].ptr),
				malloc_table[i].file, malloc_table[i].line);
}

void *debug_calloc(size_t nmemb, size_t size, const char *source_file, int source_line)
{
	void *ptr = default_calloc(nmemb, size);

	++malloc_count;
	register_ptr(ptr, source_file, source_line);

	return ptr;
}

void debug_free(void *ptr, const char *source_file, int source_line)
{
	/* See checking_free for rationale of this abort.  We repeat it here
	   because we can print the file and the line where the offending
	   free occurred.  */
	if (!ptr) {
		fprintf(stderr, _("ERROR: free NULL pointer at %s:%d\n"), source_file, source_line);
		exit(EXIT_FAILURE);
	}
	if (!unregister_ptr(ptr)) {
		fprintf(stderr, _("ERROR: bad free (0x%0*lx) at %s:%d\n"), PTR_FORMAT (ptr), source_file, source_line);
		exit(EXIT_FAILURE);
	}
	++free_count;

	default_free(ptr);
}

#endif /* defined(DEBUG) || defined(_DEBUG) */

void *default_calloc(size_t nmemb, size_t size)
{
	void *ptr = calloc(nmemb, size);

	if (!ptr) {
		fprintf(stderr, _("ERROR: %s,%d: memory allocation failed (%ld bytes)\n"), __FILE__, __LINE__, (long int) size);
		exit(EXIT_FAILURE);
	}

	return ptr;
}

