/*---------------------------------------------------------------------------
 * Copyright (C) 2008-2010 - Emanuele Bovisio
 *
 * This file is part of Mulk.
 *
 * Mulk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Mulk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Mulk.  If not, see <http://www.gnu.org/licenses/>.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 *---------------------------------------------------------------------------*/

#include "defines.h"
#ifdef _WIN32
#include "../win32/getopt.h"
#else
#include <getopt.h>
#endif
#include "file_obj.h"
#include "url_list.h"
#include "string_obj.h"
#ifdef ENABLE_METALINK
#include "metalink_list.h"
#include "country_codes.h"

#define LOC_LENGTH 2
#define LOC_OPTION_DELIM ",;.:- "
#endif

#define NUM_OPTIONS ((int) (sizeof(options) / sizeof(option_t)))

#define DEFAULT_MIME_OUTPUT_DIRECTORY "data"
#define DEFAULT_FILE_OUTPUT_DIRECTORY ""
#define DEFAULT_TEMP_DIRECTORY ".tmp-mulk"


typedef enum type_option_t {
	OPTION_BOOL,
	OPTION_INT,
	OPTION_STRING,
	OPTION_EXIT,
	OPTION_VERSION
} type_option_t;

typedef struct option_t {
	char *long_opt;
	char short_opt;
	char *description;
	void *value;
	type_option_t type;
	int min_val;
	int max_val;
	char *error;
	char *title;
} option_t;

#ifdef ENABLE_RECURSION
static char **default_domains = NULL;
static char **accepted_domains = NULL;
static char **rejected_domains = NULL;
#endif /* ENABLE_RECURSION */
#ifdef ENABLE_METALINK
static char *location_countries = NULL;
static char *location_continents = NULL;
#endif

/* reset all struct */
option_value_t option_values = {0};

option_t options[] = {
	{"version",                  0, gettext_noop("display version number and exit"),
		NULL,                                  OPTION_VERSION, 0, 0, NULL,
		gettext_noop("General options")},
	{"help",                   'h', gettext_noop("display this help and exit"),
		NULL,                                  OPTION_EXIT, 0, 0, NULL, NULL},
	{"quiet",                  'q', gettext_noop("quiet (no output)"),
		&option_values.quiet,                  OPTION_BOOL, 0, 0, NULL, NULL},
	{"verbose",                'v', gettext_noop("be verbose"),
		&option_values.verbosity,              OPTION_BOOL, 0, 0, NULL, NULL},
#ifdef ENABLE_RECURSION
	{"depth",                  'd', gettext_noop("maximum recursion depth"),
		&option_values.depth,                  OPTION_INT, 1, 10, gettext_noop("wrong depth values"), NULL},
#endif /* ENABLE_RECURSION */
	{"user-agent",             'U', gettext_noop("identify as different agent instead of Mulk/VERSION"),
		&option_values.user_agent,             OPTION_STRING, 0, 0, NULL, NULL},
	{"max-sim-conns-per-host", 'p', gettext_noop("maximum simultaneous connections per host"), 
		&option_values.max_sim_conns_per_host, OPTION_INT, 1, 5,
		gettext_noop("wrong value for maximum simultaneous connections per host"), NULL},
	{"max-sim-conns",          'm', gettext_noop("maximum simultaneous connections"),
		&option_values.max_sim_conns,          OPTION_INT, 1, 50,
		gettext_noop("wrong value for maximum simultaneous connections"), NULL},
#ifdef ENABLE_RECURSION
	{"span-hosts",             'H', gettext_noop("go to foreign hosts"),
		&option_values.span_hosts,             OPTION_BOOL, 0, 0, NULL, NULL},
	{"domains",                'D', gettext_noop("comma-separated list of accepted domains"),
		&option_values.domains,                OPTION_STRING, 0, 0, NULL, NULL},
	{"exclude-domains",          0, gettext_noop("comma-separated list of rejected domains"),
		&option_values.exclude_domains,        OPTION_STRING, 0, 0, NULL, NULL},
#endif /* ENABLE_RECURSION */
	{"user",                     0, gettext_noop("username to use for the connection"),
		&option_values.user,                   OPTION_STRING, 0, 0, NULL, NULL},
	{"password",                 0, gettext_noop("password to use for the connection"),
		&option_values.password,               OPTION_STRING, 0, 0, NULL, NULL},
#ifdef ENABLE_RECURSION
	{"follow-ftp",               0, gettext_noop("follow FTP links from HTML documents"),
		&option_values.follow_ftp,             OPTION_BOOL, 0, 0, NULL, NULL},
#endif /* ENABLE_RECURSION */
	{"option-file",            't', gettext_noop("text file with list of options"),
		&option_values.option_filename,        OPTION_STRING, 0, 0, NULL,
		gettext_noop("Input options")},
	{"url-file",               'f', gettext_noop("text file with list of URLs to download"),
		&option_values.url_filename,           OPTION_STRING, 0, 0, NULL, NULL},
#ifdef ENABLE_METALINK
	{"metalink-file",          'l', gettext_noop("Metalink filename"),
		&option_values.metalink_filename,      OPTION_STRING, 0, 0, NULL,
		gettext_noop("Metalink options")},
	{"metalink-location",        0, gettext_noop("comma-separated list of accepted locations"),
		&option_values.metalink_location,      OPTION_STRING, 0, 0, NULL, NULL},
	{"metalink-continent",       0, gettext_noop("comma-separated list of accepted continents"),
		&option_values.metalink_continent,     OPTION_STRING, 0, 0, NULL, NULL},
	{"metalink-os",              0, gettext_noop("operating system to be considered in Metalink files"),
		&option_values.metalink_os,            OPTION_STRING, 0, 0, NULL, NULL},
	{"metalink-language",        0, gettext_noop("language to be considered in Metalink files"),
		&option_values.metalink_language,      OPTION_STRING, 0, 0, NULL, NULL},
#ifdef ENABLE_CHECKSUM
	{"metalink-resume-file",     0, gettext_noop("filename to resume"),
		&option_values.metalink_resume_file,   OPTION_STRING, 0, 0, NULL, NULL},
#endif /* ENABLE_CHECKSUM */
	{"follow-metalink",          0, gettext_noop("follow Metalink files from HTML documents"),
		&option_values.follow_metalink,        OPTION_BOOL, 0, 0, NULL, NULL},
#endif /* ENABLE_METALINK */
	{"report-file",            'r', gettext_noop("report filename"),
		&option_values.report_filename,        OPTION_STRING, 0, 0, NULL,
		gettext_noop("Reporting options")},
	{"report-csv-file",        'c', gettext_noop("report CSV filename"),
		&option_values.report_csv_filename,    OPTION_STRING, 0, 0, NULL, NULL},
	{"disable-site-save",      'x', gettext_noop("don't save whole site tree to disk (enabled by default)"),
		&option_values.disable_save_tree,      OPTION_BOOL, 0, 0, NULL,
		gettext_noop("Saving options")},
	{"save-gif-image",         'g', gettext_noop("save GIF images to mime output directory"),
		&option_values.save_gif_image,         OPTION_BOOL, 0, 0, NULL, NULL},
	{"save-png-image",         'n', gettext_noop("save PNG images to mime output directory"),
		&option_values.save_png_image,         OPTION_BOOL, 0, 0, NULL, NULL},
	{"save-jpeg-image",        'j', gettext_noop("save JPEG images to mime output directory"),
		&option_values.save_jpeg_image,        OPTION_BOOL, 0, 0, NULL, NULL},
	{"save-mime-type",           0, gettext_noop("save URLs with specific mime-type to mime output directory"),
		&option_values.save_mime_type,         OPTION_STRING, 0, 0, NULL, NULL},
	{"mime-output-dir",          0, gettext_noop("mime output directory"),
		&option_values.mime_output_directory,  OPTION_STRING, 0, 0, NULL, NULL},
	{"file-output-dir",          0, gettext_noop("file output directory"),
		&option_values.file_output_directory,  OPTION_STRING, 0, 0, NULL, NULL},
	{"temp-dir",                 0, gettext_noop("temporary directory"),
		&option_values.temp_directory,         OPTION_STRING, 0, 0, NULL, NULL},
	{"min-image-width",          0, gettext_noop("minimum image width"),
		&option_values.min_image_width,        OPTION_INT, 0, INT_MAX,
		gettext_noop("wrong minimum image width"),
		gettext_noop("Image options (active only with --save-gif-image, --save-png-image or --save-jpeg-image)")},
	{"max-image-width",          0, gettext_noop("maximum image width"),
		&option_values.max_image_width,        OPTION_INT, 0, INT_MAX,
		gettext_noop("wrong maximum image width"), NULL},
	{"min-image-height",         0, gettext_noop("minimum image height"),
		&option_values.min_image_height,       OPTION_INT, 0, INT_MAX,
		gettext_noop("wrong minimum image height"), NULL},
	{"max-image-height",         0, gettext_noop("maximum image height"),
		&option_values.max_image_height,       OPTION_INT, 0, INT_MAX,
		gettext_noop("wrong maximum image height"), NULL},
};

#ifdef ENABLE_RECURSION
/* domains options */

static int is_domain_empty(char **domains)
{
	return !domains || !*domains || !**domains;
}

static int count_domains(char **domains)
{
	int i;

	if (!domains || !*domains || !**domains)
		return 0;

	for (i = 0; domains[i]; i++);

	return i;
}

int add_url_to_default_domains(UriUriA *uri)
{
	int i, dom_num;
	char **new_default_domains = NULL;

	if (!uri || !uri->hostText.first || !uri->hostText.afterLast || (uri->hostText.afterLast - uri->hostText.first) <= 0)
		return 1;

	/* yet present */
	if (is_host_equal_domains(uri, default_domains))
		return 0;

	dom_num = count_domains(default_domains);

	new_default_domains = m_calloc(dom_num + 2, sizeof(char*));

	if (default_domains) {
		for (i = 0; default_domains[i]; i++)
			new_default_domains[i] = default_domains[i];

		m_free(default_domains);
	}

	new_default_domains[dom_num] = string_nnew(uri->hostText.first, uri->hostText.afterLast - uri->hostText.first);

	default_domains = new_default_domains;

	return 0;
}

static int parse_domains(const char *dom_option, char ***domains)
{
	char *buf = NULL;
	char *s, *res;
	int dom_num, i;

	if (!domains || !dom_option || !*dom_option)
		return 1;

	if (*domains)
		m_free(*domains);

	buf = string_new(dom_option);
	s = buf;
	dom_num = 0;
	while (strtok(s, DOMAIN_OPTION_DELIM)) {
		dom_num++;
		s = NULL;
	}
	string_free(&buf);

	*domains = m_calloc(dom_num + 1, sizeof(char*));

	buf = string_new(dom_option);
	s = buf;
	i = 0;
	while ((res = strtok(s, DOMAIN_OPTION_DELIM))) {
		(*domains)[i++] = string_new(res);

		MULK_NOTE((_("Add domain to list: %s\n"), res));
		s = NULL;
	}

	string_free(&buf);
	return 0;
}

int is_host_compatible_with_domains(UriUriA *uri)
{
	if (!uri)
		return 0;

	if (is_host_equal_domains(uri, default_domains))
		return 1;

	if (!option_values.span_hosts)
		return 0;

	return (is_domain_empty(accepted_domains) || is_host_in_domains(uri, accepted_domains))
		&& !is_host_in_domains(uri, rejected_domains);
}

static void free_domain(char ***domains)
{
	int i;

	if (!domains || !*domains)
		return;

	for (i = 0; (*domains)[i]; i++)
		string_free(&(*domains)[i]);

	if (*domains) {
		m_free(*domains);
		*domains = NULL;
	}
}

static void free_domains(void)
{
	free_domain(&default_domains);
	free_domain(&accepted_domains);
	free_domain(&rejected_domains);
}
#endif /* ENABLE_RECURSION */

#ifdef ENABLE_METALINK
/* Metalink locations options */

static int save_locations(const char *loc_option, char **locations)
{
	char *buf = NULL;
	char *s, *res, *ptr;

	string_free(locations);

	if (!locations || !loc_option || !*loc_option)
		return 0;

	ptr = string_new(loc_option);
	buf = string_new(loc_option);

	s = buf;
	while ((res = strtok(s, LOC_OPTION_DELIM))) {
		if (strlen(res) != LOC_LENGTH) {
			MULK_ERROR((_("ERROR: Metalink location length is wrong: %s\n"), res));
			*locations = NULL;
			string_free(&ptr);
			string_free(&buf);
			return 1;
		}

		strcpy(ptr, res);
		ptr += LOC_LENGTH;
		MULK_NOTE((_("Add Metalink location to list: %s\n"), res));
		s = NULL;
	}

	*locations = ptr;
	string_free(&buf);
	return 0;
}

int is_location_in_list(const char *loc)
{
	char *ptr;
	country_code_t *country;

	/* if there aren't locations then the test is true */
	if ((!location_countries || !*location_countries)
	   	&& (!location_continents || !*location_continents))
		return 1;

	if (!loc || strlen(loc) != LOC_LENGTH)
		return 0;

	if (location_countries && *location_countries) {
		for (ptr = location_countries; *ptr && string_ncasecmp(ptr, loc, LOC_LENGTH); 
			ptr += LOC_LENGTH);

		/* country found */
		if (*ptr)
			return 1;
	}

	if (location_continents && *location_continents) {
		/* find continent */
		for (country = countries; country->continent_code 
			&& string_ncasecmp(country->two_letter_code, loc, LOC_LENGTH); country++);

		if (!country->continent_code)
			return 0;

		for (ptr = location_continents; *ptr && string_ncasecmp(ptr, country->continent_code, LOC_LENGTH);
			ptr += LOC_LENGTH);

		/* country found */
		if (*ptr)
			return 1;
	}

	return 0;
}

static void free_locations(void)
{
	string_free(&location_countries);
	string_free(&location_continents);
}
#endif /* ENABLE_METALINK */

/* general options */
static struct option *create_long_options_array(void)
{
	int i;
	struct option *long_opts;

	long_opts = m_calloc(NUM_OPTIONS + 1, sizeof(struct option));

	for (i = 0; i < NUM_OPTIONS; i++) {
		long_opts[i].name = options[i].long_opt;
		long_opts[i].has_arg = (options[i].type == OPTION_INT || options[i].type == OPTION_STRING)
			? required_argument : no_argument;
		long_opts[i].flag = NULL;
		long_opts[i].val = options[i].short_opt;

		MULK_DEBUG((_("long option inserted %d, %s\n"), i, long_opts[i].name));
	}

	return long_opts;
}

static char *create_short_options_string(struct option *long_opts)
{
	int i, count = 0;
	char *short_opts, *ptr;

	for (i = 0; long_opts[i].name; i++)
		if (long_opts[i].val) {
			count++;
			if (long_opts[i].has_arg == required_argument)
				count++;
		}

	short_opts = ptr = string_alloc(count);

	for (i = 0; long_opts[i].name; i++)
		if (long_opts[i].val) {
			*(ptr++) = (char) long_opts[i].val;
			if (long_opts[i].has_arg == required_argument)
				*(ptr++) = ':';
		}

	MULK_DEBUG((_("short option string %s\n"), short_opts));
	return short_opts;
}

void printf_usage(void)
{
	int i;

#ifdef ENABLE_METALINK
	printf(_("Mulk %s, a non-interactive multi-connection network downloader with image filtering and Metalink support.\n"),
		mulk_version_number);
#else
	printf(_("Mulk %s, a non-interactive multi-connection network downloader with image filtering.\n"),
		mulk_version_number);
#endif
	printf(_("Usage: Mulk [OPTION]... [URL]...\n"));

	for (i = 0; i < NUM_OPTIONS; i++) {
		if (options[i].title)
			printf("\n%s:\n", _(options[i].title));

		if (options[i].short_opt)
			printf("  -%c,", options[i].short_opt);
		else
			printf("     ");

		printf(" --%-25s%s\n", options[i].long_opt, _(options[i].description));
	}

	printf("\n");
}

int find_short_option(const char name)
{
	int i;

	if (!name)
		return -1;

	for (i = 0; i < NUM_OPTIONS && options[i].short_opt != name; i++);

	return (i < NUM_OPTIONS) ? i : -1;
}

int find_long_option(const char *name)
{
	int i;

	if (!name || !*name)
		return -1;

	for (i = 0; i < NUM_OPTIONS && strcmp(options[i].long_opt, name); i++);

	return (i < NUM_OPTIONS) ? i : -1;
}

void init_options(void)
{
#ifdef ENABLE_RECURSION
	option_values.depth = 1;
#endif
	option_values.max_sim_conns_per_host = 2;
	option_values.max_sim_conns = 50;

	option_values.mime_output_directory = string_new(DEFAULT_MIME_OUTPUT_DIRECTORY);
	option_values.file_output_directory = string_new(DEFAULT_FILE_OUTPUT_DIRECTORY);
	option_values.temp_directory = string_new(DEFAULT_TEMP_DIRECTORY);
}

int set_option(int ind, const char *value)
{
	int val;
	char *name;

	if (ind < 0 || ind >= NUM_OPTIONS) {
		fprintf(stderr, _("\nERROR: option not recognised\n\n"));
		return 1;
	}

	name = options[ind].long_opt;

	switch (options[ind].type) {
		case OPTION_BOOL:
			*((int*) options[ind].value) = 1;
			break;
		case OPTION_INT:
			if (!value) {
				fprintf(stderr, _("\nERROR: this option (%s) needs to be followed by a value\n\n"), name ? name : "");
				return 1;
			}
			val = atoi(value);
			if (val < options[ind].min_val || val > options[ind].max_val) {
				fprintf(stderr, _("\nERROR: %s\n\n"), _(options[ind].error));
				return 1;
			}
			*((int*) options[ind].value) = val;
			break;
		case OPTION_STRING:
			if (!value) {
				fprintf(stderr, _("\nERROR: this option (%s) needs to be followed by a value\n\n"), name ? name : "");
				return 1;
			}
			if (*((char**) options[ind].value))
				string_free((char**) options[ind].value);

			*((char**) options[ind].value) = string_new(value);
			break;
		case OPTION_EXIT:
			return 1;
		case OPTION_VERSION:
			printf(_("Mulk version %s\n"), mulk_version_number);
			return 2;
		default:
			fprintf(stderr, _("\nERROR: option type not valid, %d\n\n"), options[ind].type);
			return 1;
	}

	return 0;
}

int set_options(int argc, char **argv)
{
	struct option *long_options = create_long_options_array();
	char *short_options = create_short_options_string(long_options);
	int c, option_index, ret;

	for (;;) {
		option_index = 0;

		c = getopt_long(argc, argv, short_options, long_options, &option_index);

		/* option finished */
		if (c == -1)
			break;

		/* not recognised */
		if (c == '?') {
			string_free(&short_options);
			m_free(long_options);
			return 1;
		}

		/* short option */
		if (c)
			option_index = find_short_option((char) c);

		MULK_DEBUG((_("option recognised %d, %s\n"), option_index, 
			option_index >= 0 ? long_options[option_index].name : "(NULL)"));

		if ((ret = set_option(option_index, optarg)) != 0) {
			string_free(&short_options);
			m_free(long_options);
			return ret;
		}
	}

	string_free(&short_options);
	m_free(long_options);

	/* read options from a file */
	if (option_values.option_filename) 
		if ((ret = read_option_from_text_file(option_values.option_filename)) != 0)
			return ret;

#ifdef ENABLE_RECURSION
	if (option_values.domains)
		if (parse_domains(option_values.domains, &accepted_domains))
			return 1;

	if (option_values.exclude_domains)
		if (parse_domains(option_values.exclude_domains, &rejected_domains))
			return 1;
#endif /* ENABLE_RECURSION */

	if (optind >= argc && !option_values.url_filename 
#ifdef ENABLE_METALINK
		&& !option_values.metalink_filename
#endif
		) {
		fprintf(stderr, _("\nERROR: url not present\n\n"));
		return 1;
	}

	while (optind < argc) {
		MULK_NOTE((_("Add url to download coming from command line: %s\n"), argv[optind]));
		add_new_url(argv[optind++]);
	}

	if (option_values.url_filename)
		if (read_uri_from_text_file(option_values.url_filename))
			return 1;

#ifdef ENABLE_METALINK
	if (option_values.metalink_filename) {
		if (save_locations(option_values.metalink_location, &location_countries))
			return 1;
		if (save_locations(option_values.metalink_continent, &location_continents))
			return 1;

		add_new_metalink(option_values.metalink_filename, 1,
#ifdef ENABLE_CHECKSUM
			option_values.metalink_resume_file);
#else /* not ENABLE_CHECKSUM */
			NULL);
#endif /* not ENABLE_CHECKSUM */
	}
#endif /* ENABLE_METALINK */

	return 0;
}

void free_options(void)
{
	int i;

	for (i = 0; i < NUM_OPTIONS; i++)
		if (options[i].type == OPTION_STRING)
			string_free((char **) options[i].value);

#ifdef ENABLE_RECURSION
	free_domains();
#endif
#ifdef ENABLE_METALINK
	free_locations();
#endif
}

